// CRPuppet.cpp: CRPuppet NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////

#include "CRPuppet.h"

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////
namespace Spr{;

SGOBJECTIMP(CRPuppet, CRHuman);

//////////////////// PuppetSpringNX ////////////////////
inline bool IsValid(const Vec3f& v){
	return v.square() < Square(900);
}

Vec3f CRPuppet::PositionSpring::GetPos(){
	if (solid) return solid->GetFrame()->GetPosture() * pos;
	return Vec3f();
}

Vec3f CRPuppet::PositionSpring::GetVel(){
	if (solid){
		Vec3f gPos = solid->GetOrientation() * pos;
		Vec3f vel = solid->GetVelocity() + (solid->GetAngularVelocity() ^ gPos);
		return vel;
	};
	return Vec3f();
}

Vec3f CRPuppet::PositionSpring::GetForce(){
	return force;
}

void CRPuppet::PositionSpring::SetSolid(PHSolid* s, Vec3f p, float spr, float dmp){
	solid   = s;
	pos     = p;
	sprRate = spr;
	dmpRate = dmp;
}

void CRPuppet::PositionSpring::SetTarget(Vec3f pos, Vec3f vel, bool b){
	targetPos = pos;
	targetVel = vel;
	bForce   = b;
}

void CRPuppet::PositionSpring::AddForce(Vec3f f){
	force = f;
	if (solid) solid->AddForce(f, GetPos());
}

void CRPuppet::PositionSpring::AddSpringForce(float dt){
/*	const float SPRING = 0.04f * sprRate;		//	ol
	const float DAMPER = 0.9f  * dmpRate;		//	_p*/
	const float SPRING = 0.06f * sprRate;		//	ol
	const float DAMPER = 0.15f  * dmpRate;		//	_p
	dt = 0.006f;

	if(solid){
		float mass = solid->GetMass();
		float spring = SPRING * mass * 2 / (dt*dt);
		float damper = DAMPER * mass / (dt);

		Vec3f pos2 = GetPos(); // ݂Pos,Vel
		Vec3f vel2 = GetVel();

		if(IsValid(targetPos) && bForce){
			Vec3f dPos = targetPos - pos2;
			Vec3f dVel = targetVel - vel2;
			if (dVel.norm() > 1e5f) dVel = dVel.unit() * 1e5f;
			if (dPos.norm() > 1e2f) dPos = dPos.unit() * 1e2f;
			force = spring * dPos + damper * dVel;

			if (force.norm() / mass > 1e3f) force = force.unit() * mass * 1e3f;

			AddForce(force);
		}
	}
	bForce = false;
}

//////////////////// PostureSpringsNX ////////////////////
void CRPuppet::PositionSprings::Draw(GRRender* render){
	render->SetModelMatrix(Affinef());
	render->SetLineWidth(10);
	for(iterator it = begin(); it != end(); ++it){
		if(it->targetPos.norm() != 0){
			render->SetMaterial(GRMaterialData(Vec4f(0,1,1,1)));
			Vec3f pos = it->targetPos;
			render->DrawDirect(GRRender::POINTS, &pos, &pos+1);
		}
	}
}

//////////////////// PostureSpringNX ////////////////////
Quaterniond CRPuppet::PostureSpring::GetQuaternion(){
	if (solid) return solid->GetOrientation();
	return Quaterniond();
}

Vec3f CRPuppet::PostureSpring::GetAngularVelocity(){
	if (solid) return solid->GetAngularVelocity();
	return Vec3f();
}

Vec3f CRPuppet::PostureSpring::GetTorque(){
	return torque;
}

void CRPuppet::PostureSpring::SetSolid(PHSolid* s, float spr, float dmp){
	solid   = s;
	sprRate = spr;
	dmpRate = dmp;
}

void CRPuppet::PostureSpring::SetTarget(Quaterniond quat, Vec3f angVel, bool b){
	targetQuat = quat;
	targetAngV = angVel;
	bTorque   = b;
}

void CRPuppet::PostureSpring::AddTorque(Vec3f t){
	torque = t;
	if (solid) solid->AddTorque(t);
}

void CRPuppet::PostureSpring::AddSpringTorque(float dt){
/*	const float SPRING = 0.008f * sprRate;		//	ol
	const float DAMPER = 1.2f * dmpRate;		//	_p*/
	const float SPRING = 2.0f * sprRate;		//	ol
	const float DAMPER = 4.0f * dmpRate;		//	_p
	dt = 0.006f;

	Matrix3d inertia = solid->GetInertia();
	Matrix3d spring = (SPRING * 2 / (dt*dt)) * inertia;
	Matrix3d damper = (DAMPER / dt) * inertia;
//	float det = solid->GetInertia().det();
//	float spring = (SPRING * 2 / (dt*dt)) * det;
//	float damper = (DAMPER / dt) * det;

	Quaterniond quat = GetQuaternion(); // ݂Quat,AngV
	Vec3f angV = GetAngularVelocity();

	if(bTorque){
		Quaterniond dQuat = targetQuat * quat.inv();
		if (dQuat.W() < 0){	// dQuat̐K
			dQuat *= -1;
		}
		Vec3f dAngV = targetAngV - angV;
		torque = spring * dQuat.rotation() + damper * dAngV;
		if(torque.norm() > 1e5f) torque = torque.unit() * 1e5f;
		AddTorque(torque);
	}
}

//////////////////// ReachingMovementNX ////////////////////
CRPuppet::ReachingMovement::ReachingMovement(){
	Init();
}

void CRPuppet::ReachingMovement::Init(){
	bForce = false;
	bActive = false;
	state = 0;
	time = 0.0f;
	targetSolid = NULL;
}

void CRPuppet::ReachingMovement::SetSpring(PHSolid* s, Vec3f r){
	solid = s;
	pos   = r;
	sprRate = 1.0f;
	dmpRate = 1.0f;
}

void CRPuppet::ReachingMovement::SetTimer(float t, float o){
	time     = t;
	offset   = o;
	bActive  = true;
}

void CRPuppet::ReachingMovement::SetTargetPos(Vec3f p, Vec3f v){
	finalPos = p;
	finalVel = v;
}

void CRPuppet::ReachingMovement::SetTargetSolid(PHSolid* so, Vec3f p, Vec3f v){
	targetSolid = so;
	localPos = p;
	finalVel = v;
}

void CRPuppet::ReachingMovement::SetType(int type){
	state = type;
}

void CRPuppet::ReachingMovement::Draw(GRRender* render){
	if(bActive){
		render->SetModelMatrix(Affinef());
		render->SetLineWidth(10);
		render->SetMaterial(GRMaterialData(Vec4f(1,0,0,1)));
		Vec3f pos[2];
		pos[0] = finalPos;
		pos[1] = GetPos();
		render->DrawDirect(GRRender::POINTS, pos, pos+1);
		render->DrawDirect(GRRender::LINES, pos, pos+2);
	}
}

void CRPuppet::ReachingMovement::Step(SGScene* scene){
	float dt = scene->GetTimeStep();
	if(bActive){
		if(time <= -offset){
			Init();
			return;
		}

		if(targetSolid) finalPos = targetSolid->GetFrame()->GetPosture() * localPos;	// ڕWʒu̍XV
		if(time > dt){
			Vec3f pos, vel;
			float s = dt / time;
			pos = (finalPos - GetPos()) * s + GetPos();	// ڕW̐ݒ
			vel = (finalPos - GetPos()) / time;
			//pos = (finalPos - GetPos()) * (10*pow(s,3) - 15*pow(s,4) + 6*pow(s,5)) + GetPos();	// ڕW̐ݒ(xŏ)
			//vel = (finalPos - GetPos()) * (pow(s,2) - 2*pow(s,3) + pow(s,4)) * 30 / time;
			SetTarget(pos, vel, true);
		}
		else{
			SetTarget(finalPos, Vec3f(), true);
		}
		AddSpringForce(dt);

		//GRRender* render;
		//scene->GetRenderers().Find(render);
		//Draw(render);

		time -= dt;
	}
}

//////////////////// LocusGeneratorNX ////////////////////
void CRPuppet::LocusGenerator::SaveLog(CRPuppet* puppet){

	for(int i = LogNum-1; i > 0; --i) logPos[0][i] = logPos[0][i-1];
	logPos[0][0] = puppet->solids[6]->GetCenterPosition();

	for(int i = LogNum-1; i > 0; --i) logPos[1][i] = logPos[1][i-1];
	logPos[1][0] = puppet->solids[9]->GetCenterPosition();

	for(int i = LogNum-1; i > 0; --i) logPos[2][i] = logPos[2][i-1];
	logPos[2][0] = puppet->solids[2]->GetCenterPosition();

	for(int i = LogNum-1; i > 0; --i) logPos[3][i] = logPos[3][i-1];
	logPos[3][0] = puppet->solids[3]->GetCenterPosition();
}

inline float SumX(int n, int p, float* y){
	float sum = 0;
	float w = 1;
	for(int i = 0; i < n; ++i){
		w = 2 * n - i;
		//w = pow(i + 1, -2);
		if(y) sum += w * pow(i, p) * y[i];
		else  sum += w * pow(i, p);
	}
	return sum;
}

void CRPuppet::LocusGenerator::WeightedLeastSquares(float* y, int n, int m, PTM::VVector<float>* a){
	// dݕtŏ@
	// n _(y[i]) m ߎ (x[i] ͓Ԋu, n > m)
	PTM::VMatrixRow<float> sum_x;
	PTM::VVector<float>    sum_y;
	sum_x.resize(m+1,m+1);
	sum_y.resize(m+1);
	sum_y.clear(0);
	sum_x.clear(0);

	for(int i = 0; i <= 2 * m; ++i){
		float sumX;
		if(i % 2 == 0) sum_y[i/2] = SumX(n, m - i/2, y);
		sumX = SumX(n, 2 * m - i, NULL);
		if(i <= m) for(int j = 0; j <= i; ++j)     sum_x[j][i-j] = sumX;
		else       for(int j = i - m; j <= m; ++j) sum_x[j][i-j] = sumX;
	}
	*a = sum_x.inv() * sum_y;
}

void CRPuppet::LocusGenerator::CalFutureLocus(Vec3f* past, int n, int m, Vec3f* a){
	// ߋ_ŏ@łߎč̋OՂ߂
	Vec3f p = Vec3f();
	float *y;
	PTM::VVector<float> ai;
	ai.resize(m+1);
	y  = (float *) calloc(n, sizeof(float));

	for(int i = 0; i < n; ++i) y[i] = past[i].X();
	WeightedLeastSquares(y, n, m, &ai);
	for(int i = 0; i <= m; ++i) a[i].X() = ai[i];

	for(int i = 0; i < n; ++i) y[i] = past[i].Y();
	WeightedLeastSquares(y, n, m, &ai);
	for(int i = 0; i <= m; ++i) a[i].Y() = ai[i];

	for(int i = 0; i < n; ++i) y[i] = past[i].Z();
	WeightedLeastSquares(y, n, m, &ai);
	for(int i = 0; i <= m; ++i) a[i].Z() = ai[i];

	free(y);
}

Vec3f CRPuppet::LocusGenerator::CalcFuturePosition(Vec3f* a, int t){
	Vec3f pos = Vec3f();
	for(int i = 0; i <= Deg; ++i) pos += a[i] * pow(-t, Deg-i);
	return pos;
}

Vec3f CRPuppet::LocusGenerator::CalcFutureVelocity(Vec3f*a, int t, float dt){
	// OՂߎAXebv t ł̑xvZ
	Vec3f vel = Vec3f();
	for(int i = 0; i <= Deg; ++i) vel += a[i] * (pow(-(t+0.5), Deg-i) - pow(-(t-0.5), Deg-i));
	return vel / dt;
}

void CRPuppet::LocusGenerator::DrawFutureLocus(Vec3f*a, int t, int k, GRRender* render, Vec4f color){
	// OՂߎAt * k Xbev̈ʒu\
	render->SetModelMatrix(Affinef());
	render->SetLineWidth(10);
	for(int i = 0; i <= k; ++i){
		color.w = 1-(float)i/k;
		render->SetMaterial(GRMaterialData(color));
		Vec3f pos = Vec3f();
		for(int j = 0; j <= Deg; ++j) pos += a[j] * pow(-t*i, Deg-j);
		render->DrawDirect(GRRender::POINTS, &pos, &pos+1);
	}
}

void CRPuppet::LocusGenerator::SetLocusCoefficient(CRPuppet* puppet, SGScene* scene){
	GRRender* render;
	scene->GetRenderers().Find(render);

	for(int i = 0; i < 2; ++i){
		CalFutureLocus(logPos[i], LogNum, Deg, ak[i]);
		//DrawFutureLocus(ak[i], 10, 5, render, Vec4f(0,1,1,1));
	}
}

//////////////////// CRPuppetNX ////////////////////
CRPuppet::CRPuppet(){

}

void CRPuppet::LoadDerivedModel(SGScene* scene){
	if(IsLoaded()){
		ChangeJointRange();
		SetJointBasicPos();
		SetSprings();
		for(int i = 0; i < 3; ++i) rm[i].Init();
	}
}

bool CRPuppet::Connect(UTRef<SGScene> scene){
	solids.clear();
	joints.clear();
	jointPids.clear();

	SetLoaded(true);

	// Solid Connect
	ConnectSolid("soWaist", scene);
	ConnectSolid("soAbdomen", scene);
	ConnectSolid("soChest", scene);
	ConnectSolid("soHead", scene);

	ConnectSolid("soRUArm", scene);
	ConnectSolid("soRFArm", scene);
	ConnectSolid("soRHand", scene);

	ConnectSolid("soLUArm", scene);
	ConnectSolid("soLFArm", scene);
	ConnectSolid("soLHand", scene);
/*
//	ConnectSolid("soRThigh", scene);
	solids.push_back(NULL);
//	ConnectSolid("soRLThigh", scene);
	solids.push_back(NULL);
//	ConnectSolid("soRHeel", scene);
	solids.push_back(NULL);
//	ConnectSolid("soRToe", scene);
	solids.push_back(NULL);
//	ConnectSolid("soLThigh", scene);
	solids.push_back(NULL);
//	ConnectSolid("soLLThigh", scene);
	solids.push_back(NULL);
//	ConnectSolid("soLHeel", scene);
	solids.push_back(NULL);
//	ConnectSolid("soLToe", scene);
	solids.push_back(NULL);
*/
	//Joint Connect	
	ConnectJoint("joWaist1", scene);
	ConnectJoint("joWaist2", scene);
//	ConnectJoint("joWaist3", scene);
	joints.push_back(NULL);
//	ConnectJoint("joChest1", scene);
	joints.push_back(NULL);
//	ConnectJoint("joChest2", scene);
	joints.push_back(NULL);
	ConnectJoint("joChest3", scene);
	ConnectJoint("joNeck1", scene);
	ConnectJoint("joNeck2", scene);
//	ConnectJoint("joNeck3", scene);
	joints.push_back(NULL);
	
	ConnectJoint("joRShoulder1", scene);
	ConnectJoint("joRShoulder2", scene);
	ConnectJoint("joRShoulder3", scene);
	ConnectJoint("joRElbow1", scene);
//	ConnectJoint("joRElbow2", scene);
	joints.push_back(NULL);
	ConnectJoint("joRWrist1", scene);
//	ConnectJoint("joRWrist2", scene);
	joints.push_back(NULL);

	ConnectJoint("joLShoulder1", scene);
	ConnectJoint("joLShoulder2", scene);
	ConnectJoint("joLShoulder3", scene);
	ConnectJoint("joLElbow1", scene);
//	ConnectJoint("joLElbow2", scene);
	joints.push_back(NULL);
	ConnectJoint("joLWrist1", scene);
//	ConnectJoint("joLWrist2", scene);
	joints.push_back(NULL);
/*
//	ConnectJoint("joRHip1", scene);
	joints.push_back(NULL);
//	ConnectJoint("joRHip2", scene);
	joints.push_back(NULL);
//	ConnectJoint("joRHip3", scene);
	joints.push_back(NULL);
//	ConnectJoint("joRKnee", scene);
	joints.push_back(NULL);
//	ConnectJoint("joRAnkle1", scene);
	joints.push_back(NULL);
//	ConnectJoint("joRAnkle2", scene);
	joints.push_back(NULL);
//	ConnectJoint("joRAnkle3", scene);
	joints.push_back(NULL);
//	ConnectJoint("joRToe", scene);
	joints.push_back(NULL);

//	ConnectJoint("joLHip1", scene);
	joints.push_back(NULL);
//	ConnectJoint("joLHip2", scene);
	joints.push_back(NULL);
//	ConnectJoint("joLHip3", scene);
	joints.push_back(NULL);
//	ConnectJoint("joLKnee", scene);
	joints.push_back(NULL);
//	ConnectJoint("joLAnkle1", scene);
	joints.push_back(NULL);
//	ConnectJoint("joLAnkle2", scene);
	joints.push_back(NULL);
//	ConnectJoint("joLAnkle3", scene);
	joints.push_back(NULL);
//	ConnectJoint("joLToe", scene);
	joints.push_back(NULL);
*/
	for(int i=0; i<joints.size(); ++i){
		jointPids.push_back(PHJointPid::Find((PHJoint1D*)joints[i], scene));
	}
	return IsLoaded();
}

inline float GetChildMass(PHJointBase* j){
	if (strlen(j->solid->GetName()) != 0){
		return j->solid->GetMass();
	}
	else{
		float rv = 0;
		for(unsigned i = 0; i < j->Children().size(); ++i){
			rv += GetChildMass(j->Children()[i]);
		}
		return rv;
	}
}

void CRPuppet::SetJointSpring(float dt){
/*	//const float SAFETYRATE = 0.002f;
	const float SAFETYRATE = 0.002f;
	float k = 0.1f * SAFETYRATE;
	float b = 0.8f * SAFETYRATE;*/
	const float SAFETYRATE = 0.01f;
	float k = 0.6f * SAFETYRATE;
	float b = 0.8f * SAFETYRATE;
	dt = 0.006f;

	for(int i=0; i<joints.size(); ++i){
		if(joints[i] != NULL){
			float mass = GetChildMass(joints[i]);
			jointPids[i]->proportional = k * 2 * mass / (dt*dt);
			jointPids[i]->differential = b * mass / dt;
			jointPids[i]->integral = k * 2 * mass / (dt*dt) / 5000.0f;
		}
	}
	// ֐߂_炩߂ɐݒ(㔼g)
	for(int i = 0; i < 9; ++i){ 
		if(jointPids[i] != NULL){
			JointPIDMul(jointPids[i], 0.8f, 1.0f);
		}
	}
	// ֐߂_炩߂ɐݒ(Er)
	for(int i = 9; i < 14; ++i){ 
		if(jointPids[i] != NULL){
			JointPIDMul(jointPids[i], 0.04f, 0.8f);
		}
	}
	// ֐߂_炩߂ɐݒ(r)
	for(int i = 16; i < 21; ++i){ 
		if(jointPids[i] != NULL){
			JointPIDMul(jointPids[i], 0.04f, 0.8f);
		}
	}
}

void CRPuppet::ChangeJointRange(){
	SetOneJointRange((PHJoint1D*)joints[5], -90, 90);
}

void CRPuppet::SetJointBasicPos(){
	if(jointPids[0])  jointPids[0]->goal  = jinfo[0].initPos  = 0.2f;
	if(jointPids[9])  jointPids[9]->goal  = jinfo[9].initPos  = 0.8f;
	if(jointPids[10]) jointPids[10]->goal = jinfo[10].initPos = 0.0f;
	if(jointPids[11]) jointPids[11]->goal = jinfo[11].initPos = 0.5f;
	if(jointPids[12]) jointPids[12]->goal = jinfo[12].initPos = 2.5f;

	if(jointPids[16]) jointPids[16]->goal = jinfo[16].initPos = 1.2f;
	if(jointPids[17]) jointPids[17]->goal = jinfo[17].initPos = 0.0f;
	if(jointPids[18]) jointPids[18]->goal = jinfo[18].initPos = 0.5f;
	if(jointPids[19]) jointPids[19]->goal = jinfo[19].initPos = 2.3f;
}

void CRPuppet::Draw(GRRender* render){
	positionSprings.Draw(render);
}

void CRPuppet::DrawRM(GRRender* render){
	render->SetModelMatrix(Affinef());
	render->SetLineWidth(10);
	for(int i = 0; i < 3; ++i){
		if(rm[i].bActive){
			render->SetMaterial(GRMaterialData(Vec4f(0,1,1,1)));
			Vec3f pos = rm[i].targetPos;
			render->DrawDirect(GRRender::POINTS, &pos, &pos+1);
		}
	}
}

void CRPuppet::SetSprings(){
	if(!IsLoaded()) return;
	positionSprings.clear();

	// [0] ({ʒuɌŒ)
	//postureSpring.SetSolid(solids[0], 1.0f, 1.0f);
	postureSpring.SetSolid(solids[0], 0.5f, 1.0f);
	PositionSpring positionSpr;
	//positionSpr.SetSolid(solids[0], Vec3f(0, 0, 0), 1.0f, 1.0f);
	positionSpr.SetSolid(solids[0], Vec3f(0, 0, 0), 0.1f, 1.0f);
	positionSprings.push_back(positionSpr);
}

void CRPuppet::SetFixedPos(){
	if(!IsLoaded()) return;

	postureSpring.SetTarget(Quaterniond(0,0,1,0), Vec3f(), true);
	positionSprings[0].SetTarget(Vec3f(0.0f, 1.0f,-1.0f), Vec3f(), true);
}

void CRPuppet::SetExpectedPos(float dt){
	for(int i = 0; i < positionSprings.size(); ++i) 
		positionSprings[i].targetPos += positionSprings[i].targetVel * dt;
}

void CRPuppet::SetSpringForce(float dt){
	SetFixedPos();
	for(int i = 0; i < positionSprings.size(); ++i){
		positionSprings[i].AddSpringForce(dt);
	}
	postureSpring.AddSpringTorque(dt);
}

void CRPuppet::Step(SGScene* scene){
	SetSpringForce(scene->GetTimeStep());
	for(int i = 0; i < 3; ++i) rm[i].Step(scene);
}

DEF_RECORD(XPuppet,{
	GUID Guid(){ return WBGuid("1EA16CFD-B12D-479c-8F18-80AFB33012FE"); }
});
class CRPuppetLoader : public FIObjectLoader<CRPuppet>{
public:
	CRPuppetLoader(){
		UTRef<FITypeDescDb> db = new FITypeDescDb;
		db->SetPrefix("X");
		db->REG_FIELD(FLOAT);
		db->REG_RECORD_PROTO(XPuppet);
	}
	bool LoadData(FILoadScene* ctx, CRPuppet* h){
		return true;
	}
};
class CRPuppetSaver : public FIObjectSaver<CRPuppet>{
protected:
	void SaveData(class FISaveScene* ctx, FIDocNodeBase* doc, CRPuppet* h){
	}
};

DEF_REGISTER_BOTH(CRPuppet);


}