#ifndef PHSOLID_H
#define PHSOLID_H

#include <FileIO/FIDocScene.h>
#include <SceneGraph/SGScene.h>
#include <Base/TQuaternion.h>

namespace Spr{;

class SGFrame;

enum PHIntegrationMode{
		PHINT_NONE,			//ϕȂ
		PHINT_ARISTOTELIAN,	//f = mv
		PHINT_EULER,		//IC[@
		PHINT_SIMPLETIC,
		PHINT_RUNGEKUTTA2,	//QQNb^@
		PHINT_RUNGEKUTTA4	//SQNb^@
};

///	
class PHSolid : public SGObject{
	Vec3d	_angvel[4];			///<	lϕW
	Vec3d	_angacc[4];
protected:
	double		mass;			///<	
	Matrix3d	inertia;		///<	e\
	Matrix3d	inertia_inv;
	Vec3d		force;			///<					(World)
	Vec3d		torque;			///<	gN			(World)
	Vec3d		velocity;		///<	x			(World)
	Vec3d		angVelocity;	///<	px			(World)
	Vec3d		center;			///<	ʒS̈ʒu	(Local..frameposturen)
	Quaterniond quat;			///<				(World)

	/**	ʒup\t[DʂWorld̒̃t[w肷D
		Kw̃t[łȂ΂ȂȂD
		̃t[ɂ́CXP[OĂ͂ȂȂD	*/
	UTRef<SGFrame> frame;

	///	ϕ
	PHIntegrationMode integrationMode;

	//IC[̉^
	//s͑ΊpsOD
	Vec3d	Euler(const Matrix3d& I, const Vec3d& t, const Vec3d& w){
		return Vec3d(
			(t[0] - (I[2][2] - I[1][1]) * w.Y() * w.Z()) / I[0][0],
			(t[1] - (I[0][0] - I[2][2]) * w.Z() * w.X()) / I[1][1],
			(t[2] - (I[1][1] - I[0][0]) * w.X() * w.Y()) / I[2][2]);
	}
public:
	SGOBJECTDEF(PHSolid);
	PHSolid();											///< \z
	
	bool		AddChildObject(SGObject* o, SGScene* s);///< [hɎgpD
	size_t		NReferenceObjects();					///< 1
	SGObject*	ReferenceObject(size_t i);				///< t[ԂD
	void		Loaded(SGScene* scene);					///< [hȈ
	void		Step(double dt);						///<  dt i߂D
	
	void		AddForce(Vec3d f);						///< ͂ʒSɉ
	void		AddTorque(Vec3d t){ torque += t; }		///< gN
	void		AddForce(Vec3d f, Vec3d r);				///< ͂ ʒur(Worldn) ɉ
	//邵GravityEnginegĂȂ̂Ŕp
	//void		AddForceLocal(Vec3d f, Vec3d r);		///< ͂ ʒur(Localn) ɉ
	void		ClearForce();							///< ͂ƃgNNA
	Vec3d		GetForce(){return force;}				///< ꂽ
	Vec3d		GetTorque(){return torque;}				///< ꂽgN
	void		SetForce(Vec3d f){force = f;}			///< ͂ݒ肷
	void		SetTorque(Vec3d t){torque = t;}			///< gNZbg

	SGFrame*	GetFrame(){ return frame; }				///< t[̎擾
	void		SetFrame(SGFrame* f){ frame = f; }		///< t[̐ݒ @see frame
	double		GetMass(){return mass;}					///< 
	double		GetMassInv(){return 1.0 / mass;}		///< ʂ̋t
	void		SetMass(double m){mass = m;}			///< ʂ̐ݒ
	void		SetMassInv(double minv){mass = 1.0 / minv;}
														///< ʂ̋t̐ݒ
	Matrix3d	GetInertia(){return inertia;}			///< e\
	Matrix3d	GetInertiaInv(){return inertia_inv;}	///< e\̋t
	void		SetInertia(const Matrix3d& I){			///< e\̐ݒ
		inertia = I;
		inertia_inv = I.inv();
	}
	void		SetInertiaInv(const Matrix3d& Iinv){	///< e\tŐݒ
		inertia_inv = Iinv;
		//inertia = Iinv.inv();
	}

	///	ϕ̎擾
	PHIntegrationMode GetIntegrationMode(){ return integrationMode; }
	///	ϕ̐ݒ
	void SetIntegrationMode(PHIntegrationMode m){ integrationMode=m; }

	Vec3d		GetFramePosition(){return frame->GetPosition();}
	void		SetFramePosition(const Vec3d& p){frame->SetPosition(p);}
	Vec3d		GetCenterPosition(){return frame->GetPosture()*center;}
														///< dSʒu̎擾
	void		SetCenterPosition(const Vec3d& p){		///< dSʒu̐ݒ
		frame->SetPosition(p - frame->GetRotation()*center);
	}

	///	̎擾
	Matrix3d	GetRotation(){ Matrix3d rv; quat.to_matrix(rv); return rv; }
	///	̐ݒ
	void		SetRotation(const Matrix3d& r){ quat.from_matrix(r); }

	///	̎擾
	Quaterniond GetOrientation(){return quat;}
	///	̐ݒ
	void		SetOrientation(const Quaterniond& q){ quat = q; }

	///	ʒS̑x̎擾
	Vec3d		GetVelocity(){return velocity;}
	///	ʒS̑x̐ݒ
	void		SetVelocity(const Vec3d& v){velocity = v;}

	///	px̎擾
	Vec3d		GetAngularVelocity(){return angVelocity;}
	///	px̐ݒ
	void		SetAngularVelocity(const Vec3d& av){angVelocity = av;}

	///	[Jt[猩Ĉ̎ʒSʒu̐ݒ
	Vec3d		GetCenter(){return center;}
	///	[Jt[猩Ĉ̎ʒSʒu̎擾
	void		SetCenter(const Vec3d& c){center = c;}		
	///	pxt[ɔf
	void UpdateFrame();
};

class PHSolids:public std::vector< UTRef<PHSolid> >{
public:
	UTRef<PHSolid>* Find(const UTRef<PHSolid>& s){
		iterator it = std::find(begin(), end(), s);
		if (it == end()) return NULL;
		else return &*it;
	}
	UTRef<PHSolid>* Find(const UTRef<PHSolid>& s) const {
		return ((PHSolids*)this)->Find(s);
	}
};
/**	SolidێNXDSolid̍XVsD	*/
class PHSolidContainer:public SGBehaviorEngine{
	SGOBJECTDEF(PHSolidContainer);
public:
	PHSolids solids;
	bool AddChildObject(SGObject* o, SGScene* s);
	///
	int GetPriority() const {return SGBP_SOLIDCONTAINER;}
	///	xʒuAxx̐ϕ
	virtual void Step(SGScene* s);
	virtual void Loaded(SGScene* scene);
	virtual void Clear(SGScene* s){ solids.clear(); }
	///	LĂsolid̐
	virtual size_t NChildObjects(){ return solids.size(); }
	///	LĂsolid
	virtual SGObject* ChildObject(size_t i){ return solids[i]; }

	///	Ԃ̓ǂݏo
	virtual void LoadState(const SGBehaviorStates& states);
	///	Ԃ̕ۑ
	virtual void SaveState(SGBehaviorStates& states) const;
};

/**	Solid̗͂NANX	*/
class PHSolidClearForce:public SGBehaviorEngine{
	SGOBJECTDEF(PHSolidClearForce);
public:
	UTRef<PHSolidContainer> cont;
	///	NA
	virtual void Step(SGScene* s);
	virtual int GetPriority() const { return SGBP_CLEARFORCE; }
};

struct SolidInfo{
	float mass;				///<	
	Matrix3f inertia;		///<	[g
	Vec3f velocity;			///<	x
	Vec3f angularVelocity;	///<	px
	Vec3f center;			///<	Sʒu
};

}	//	namespace Spr
#endif
